"""Global variables associated to the test suite."""

#: List of CPP variables that should be defined in config.h in order to enable this suite.
need_cpp_vars = [
]

#: List of keywords that are automatically added to all the tests of this suite.
keywords = [
]

#: List of input files
inp_files = [
    "tatdep1_1.abi",
    "tatdep1_2.abi",
    "tatdep1_3.abi",
    "tatdep1_4.abi",
    "tatdep1_5.abi",
]
